import unittest

import os
from config_probe import probe, fake_probe
from config_probe.exceptions import ConfigNotFound
from hamcrest import is_, assert_that


class TestConfigProbe(unittest.TestCase):
    def test_single_file(self):
        config = probe(path=_dir("single-file"),
                       patterns=["stuff.yaml"])

        assert_that(config.key, is_("stuff-value"))
        assert_that(config.fruits, is_(["Apple", "Orange"]))

    def test_single_file_with_namespace(self):
        config = probe(path=_dir("single-file-with-namespace"),
                       patterns=["(*).json"])

        assert_that(config.stuff.key, is_("stuff-value"))

    def test_single_file_with_namespace_with_wrong_key(self):
        config = probe(path=_dir("single-file-with-namespace"),
                       patterns=["(*).json"])

        with self.assertRaises(ConfigNotFound):
            print(config.stuff.key_inexistant)

    def test_single_file_multiple_level_raising_or_not(self):
        config = probe(path=_dir("multi-level-files"),
                       patterns=["(*)/(*).yaml", "(*)/subdir/(*).yaml"])
        assert_that(config.ns1.stuff.we.need.more.cowbell, is_("ok"))

        with self.assertRaises(ConfigNotFound):
            print(config.ns1.stuff.we.need.less.cowbell)

    def test_two_files_with_subdir_namespace(self):
        config = probe(path=_dir("two-files-with-subdir-namespace"),
                       patterns=["(*)/(*).yaml"])

        assert_that(config.ns1.stuff.key1, is_("stuff from ns1"))
        assert_that(config.ns2.stuff.key2, is_("stuff from ns2"))

    def test_only_starred_parts_are_namespaced(self):
        config = probe(path=_dir("two-files-with-subdir-namespace"),
                       patterns=["(*)/stuff.yaml"])

        assert_that(config.ns1.key1, is_("stuff from ns1"))
        assert_that(config.ns2.key2, is_("stuff from ns2"))

    def test_using_only_a_star_does_not_count_toward_namespacing(self):
        config = probe(path=_dir("two-files-with-subdir-namespace"),
                       patterns=["*/stuff.yaml"])

        assert_that(config.key1, is_("stuff from ns1"))
        assert_that(config.key2, is_("stuff from ns2"))

    def test_multiple_patterns(self):
        config = probe(path=_dir("two-files-with-subdir-namespace"),
                       patterns=["ns1/(*).yaml", "(*)/stuff.yaml"])

        assert_that(config.stuff.key1, is_("stuff from ns1"))
        assert_that(config.ns2.key2, is_("stuff from ns2"))

    def test_multiple_patterns_on_same_namespaces_should_merge_recursively(self):
        config = probe(path=_dir("multi-level-files"),
                       patterns=["(*)/(*).yaml", "(*)/subdir/(*).yaml"])

        assert_that(config.ns1.stuff.content1.key1, is_("value1"))
        assert_that(config.ns1.stuff.content2.key2, is_("value2"))

    def test_pattern_order_defines_which_files_have_the_authority(self):
        config = probe(path=_dir("key-override"),
                       patterns=["file1.yaml", "file2.yaml"])
        assert_that(config.key, is_("value2"))

        config = probe(path=_dir("key-override"),
                       patterns=["file2.yaml", "file1.yaml"])
        assert_that(config.key, is_("value1"))

    def test_yaml_dict_are_merged_list_arent(self):
        config = probe(path=_dir("dict-merge"),
                       patterns=["file1.yaml", "file2.yaml"])
        assert_that(config.mydict.common_key, is_("value2"))
        assert_that(config.mydict.only_in_1, is_("value1"))
        assert_that(config.mydict.only_in_2, is_("value2"))
        assert_that(config.mydict.subdict.common_list, is_(["a2", "b2"]))
        assert_that(config.mydict.subdict.only_in_1, is_("value1"))
        assert_that(config.mydict.subdict.only_in_2, is_("value2"))

        config = probe(path=_dir("dict-merge"),
                       patterns=["file2.yaml", "file1.yaml"])
        assert_that(config.mydict.common_key, is_("value1"))
        assert_that(config.mydict.only_in_1, is_("value1"))
        assert_that(config.mydict.only_in_2, is_("value2"))
        assert_that(config.mydict.subdict.common_list, is_(["a1", "b1"]))
        assert_that(config.mydict.subdict.only_in_1, is_("value1"))
        assert_that(config.mydict.subdict.only_in_2, is_("value2"))

    def test_support_for_empty_files(self):
        probe(path=_dir("empty-files"), patterns=["*.*"])

    def test_fake_probe(self):
        config = fake_probe({
            "key": "value",
            "key2": [
                {"hey": "ho"}
            ]
        })

        assert_that(config.key, is_("value"))
        assert_that(config.key2[0].hey, is_("ho"))


def _dir(name):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), name)



import sys

from config_probe.exceptions import ConfigNotFound


class MunchWrapper(object):
    def __init__(self, munch):
        self.munch = munch

    def __getattr__(self, item):
        try:
            result = self.munch.__getattr__(item)
        except AttributeError as e:
            raise ConfigNotFound(e)
        if type(result) in _get_primitive_type():
            return result
        return MunchWrapper(self.munch.__getattr__(item))


if sys.version_info < (3,):
    def _get_primitive_type():
        return int, float, long, bool, list, str, unicode
else:
    def _get_primitive_type():
        return int, float, bool, list, str



import glob
import json

import os

import collections
import yaml
from munch import munchify
from config_probe.munch_wrapper import MunchWrapper

NAMESPACE_PLACEHOLDER = "(*)"


def probe(path, patterns):
    config = {}

    for pattern in patterns:

        glob_pattern = pattern.replace(NAMESPACE_PLACEHOLDER, "*")
        for config_file in glob.glob(os.path.join(path, glob_pattern)):
            relevant_part_of_the_path = config_file[len(path) + 1:]
            path_parts, ext = os.path.splitext(relevant_part_of_the_path)
            path_matchers, _ = os.path.splitext(pattern)

            namespaces = _deduce_namespaces(path_matchers, path_parts)

            with open(config_file) as f:
                new_values = _parsers[ext](f)

            _add_to_configuration(config, namespaces, new_values)

    return MunchWrapper(munchify(config))


def fake_probe(content):
    return munchify(content)


_parsers = {
    ".yaml": lambda f: yaml.load(f) or {},
    ".json": lambda f: json.load(f),
}


def _deduce_namespaces(path_matchers, path_parts):
    namespaces = []
    path_parts, current_part = os.path.split(path_parts)
    path_matchers, matcher = os.path.split(path_matchers)
    while current_part is not "":
        if matcher == NAMESPACE_PLACEHOLDER:
            namespaces.append(current_part)

        path_parts, current_part = os.path.split(path_parts)
        path_matchers, matcher = os.path.split(path_matchers)
    return reversed(namespaces)


def _add_to_configuration(config, namespaces, new_values):
    current_level = config
    for ns in namespaces:
        if ns not in current_level:
            current_level[ns] = {}
        current_level = current_level[ns]

    _update(current_level, new_values)


def _update(config, values):
    for k, v in values.items():
        if k in config and isinstance(v, collections.Mapping):
            _update(config[k], v)
        else:
            config[k] = v
